# -*- coding: utf-8 -*-
"""
Created on 2020/8/17
@project: SPAIC
@filename: Visualization
@author: Hong Chaofei
@contact: hongchf@gmail.com

@description: 
"""
from spaic.Network.Assembly import NetModule

class Visualization(NetModule):

    def __init__(self):
        super(Visualization, self).__init__()
