# -*- coding: utf-8 -*-
"""
Created on 2020/8/17
@project: SPAIC
@filename: Library
@author: Hong Chaofei
@contact: hongchf@gmail.com

@description: 
"""
from spaic.Network.Assembly import BaseModule
from spaic.Network.Network import Network


class NetLibrary(BaseModule):

    def __init__(self):
        super(NetLibrary, self).__init__()

        pass

